{=================================================================
 Copyright  Brendan V. Delumpa 1996 All Rights Reserved
 Program      : Many to One Copy
 Author       : Brendan V. Delumpa
 Created      : December 20, 1996
 Description  : This is a demonstration of doing copying of many
                files into one big file, much like Copy/B. However,
                this is highly efficient in that it uses two (2)
                threads to accomplish the task. In order to do this,
                I had to implement critical sections so that the
                threads would not contend for the same destination
                file. See the TCopyThr.Execute method below.
 =================================================================}
unit MultiCopy;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, Buttons, StdCtrls, FileCtrl;

type
  TMasterThr = class(TThread)
  private
    FDest : String;
  protected
    procedure Execute; override;
  public
    constructor Create(DestinationFile : String);
    function WaitForMultipleThreads(const ThrArr : array of TThread;
                                    TimeOutVal : DWORD) : Word;
  end;

type
  TCopyThr = class(TThread)
  private
    FDest     : String;
  protected
    procedure Execute; override;
  public
    constructor Create(Destination : String);
  end;

type
  TForm2 = class(TForm)
    Edit3: TEdit;
    Button1: TButton;
    Label1: TLabel;
    DriveComboBox1: TDriveComboBox;
    DirectoryListBox1: TDirectoryListBox;
    FileListBox1: TFileListBox;
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form2    : TForm2;
  CritSect : TRTLCriticalSection;
  Files    : TStrings;
  NoFiles  : Boolean;

implementation

{$R *.DFM}

{TMasterThr Code}

constructor TMasterThr.Create(DestinationFile : String);
begin
  FDest := DestinationFile;
  FreeOnTerminate := True;
  inherited Create(False);
end;

procedure TMasterThr.Execute;
var
  thr1,
  thr2 : TCopyThr;
  I    : Integer;
begin

  {Create the threads and immediately pass
   their ThreadID's into the handle array.
   Do this immediately after creating the thread
   instead of doing it enmasse just in case
   a thread finishes before being loaded into
   the array down the line. It's safer this way
   because then you have a valid ID to work with.}
  thr1 := TCopyThr.Create(FDest);
  thr2 := TCopyThr.Create(FDest);

  {Wait for all the threads to complete. Note that
   this step is really not necessary. But if you were
   to add code below this that you didn't want to execute
   before all the threads were complete, then this would be
   a temporary gatekeeper. Also, this prevents the user
   from preemptively closing the dialog before all the
   threads were complete. }
  WaitForMultipleThreads([thr1, thr2], INFINITE);
end;

function TMasterThr.WaitForMultipleThreads(const ThrArr : array of TThread;
                                           TimeOutVal : DWORD) : Word;
var
  Hndls : TWOHandleArray;
  I     : Integer;
begin
  //This loop takes all the values from the input array and
  //assigns them to the handle array (TWOHandleArray) to be
  //passed to WaitForMultipleObjects
  for I := 0 to High(ThrArr) do
    Hndls[I] := ThrArr[I].Handle;
  Result := WaitForMultipleObjects(High(ThrArr) + 1, @Hndls, True, TimeOutVal)
end;


{TCopyThr Code}


constructor TCopyThr.Create(Destination : String);
begin
  FDest := Destination;
  FreeOnTerminate := True;
  inherited Create(False);
end;

procedure TCopyThr.Execute;
var
  FileName : String;
  sStream,
  dStream : TFileStream;
  pBuf    : Pointer;
  cnt,
  bufSize : LongInt;
  FName   : String;
  I : Integer;
begin
  FileName := '';

  if NOT NoFiles then
    repeat
      try
        //Go into a critical section here to make sure you can delete an item
        //okay. This is necessary because if this wasn't here, another thread could
        //come in and delete an item and reduce count to 0 before this thread has
        //chance to do it.
          if Assigned(Files) AND (Files.Count > 0) then
            begin
              EnterCriticalSection(CritSect);
                FileName := Files[0];
                Files.Delete(0);
              LeaveCriticalSection(CritSect);
            end
          else
            Break;

        if (FileName <> '') then
          sStream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite)
        else
          Break;

        bufSize := sStream.Size;
        try
          GetMem(pBuf, bufSize);
          cnt := sStream.Read(pBuf^, bufSize);

          //Here, we enter into another critical section because we're not
          //performing manipulations on the destination file. Notice that
          //in both critical section blocks, I've indented the code in
          //between to clearly show what's being protected by the critical
          //section.
          EnterCriticalSection(CritSect);
          if FileExists(FDest) then
            dStream := TFileStream.Create(FDest, fmOpenReadWrite)
          else
            dStream := TFileStream.Create(FDest, fmCreate);
          dStream.Seek(0, soFromEnd);
          cnt := dStream.Write(pBuf^, cnt);
          LeaveCriticalSection(CritSect);

        {with the finally statments, need to make sure we get rid of
         everything.}
        finally
          FreeMem(pBuf, bufSize);
          dStream.Free;
        end;
      finally
        sStream.Free;
      end;
    until
      ((NOT Assigned(Files)) OR (Files.Count = 0));

    NoFiles := True;
end;


{TForm2 Code}


procedure TForm2.FormCreate(Sender: TObject);
begin
  //This was set to a temporary directory for testing.
  //Go ahead and delete this entry.
  DirectoryListBox1.Directory := 'C:\CopyFolder';
end;

procedure TForm2.Button1Click(Sender : TObject);
var
  I : Integer;
  thr : TMasterThr;
begin
  //Load up the master string list from which
  //the threads will get their file names to copy
  Files := TStringList.Create;
  with FileListBox1 do
    for I := 0 to Items.Count - 1 do
      if Selected[I] then
        Files.Add(Items[I]);

  thr := TMasterThr.Create(Edit3.Text);
end;

initialization
  InitializeCriticalSection(CritSect);
  NoFiles := False;

finalization
  DeleteCriticalSection(CritSect);
end.
